// Copyright 2018-present Network Optix, Inc. Licensed under MPL 2.0: www.mozilla.org/MPL/2.0/

#include "device_agent.h"

#include <vector>
#include <string>
#include <mutex>
#include <chrono>
#include <ctime>
#include <thread>
#include <type_traits>

#define NX_PRINT_PREFIX (this->logUtils.printPrefix)
#include <nx/kit/debug.h>
#include <nx/kit/utils.h>
#include <nx/kit/json.h>

#include <nx/sdk/analytics/helpers/event_metadata.h>
#include <nx/sdk/analytics/helpers/event_metadata_packet.h>
#include <nx/sdk/helpers/settings_response.h>

#include "../utils.h"

#include "stub_analytics_plugin_events_ini.h"

namespace nx {
namespace vms_server_plugins {
namespace analytics {
namespace stub {
namespace events {

using namespace nx::sdk;
using namespace nx::sdk::analytics;
using namespace std::chrono;

namespace {

enum class EventIsActive
{
    yes,
    no,
};

struct EventDescriptor
{
    std::string eventTypeId;
    std::string caption;
    std::string description;
    std::string prolongedEventKey;
    EventIsActive isActive;

    EventDescriptor(
        std::string eventTypeId,
        std::string caption,
        std::string description,
        EventIsActive isActive = EventIsActive::yes,
        std::string prolongedEventKey = "")
        :
        eventTypeId(std::move(eventTypeId)),
        caption(std::move(caption)),
        description(std::move(description)),
        prolongedEventKey(prolongedEventKey),
        isActive(isActive)
    {
    }
};

static const std::vector<EventDescriptor> kEventsToFire = {
    {
        kObjectInTheAreaEventType,
        "Object in the Area - prolonged event (caption) key1, STARTED",
        "Object in the Area - prolonged event (description) key1, STARTED",
        EventIsActive::yes,
        "key1"
    },
    {
        kObjectInTheAreaEventType,
        "Object in the Area - prolonged event (caption) DUPLICATE key1, STARTED",
        "Object in the Area - prolonged event (description) DUPLICATE key1, STARTED",
        EventIsActive::yes,
        "key1"
    },
    {
        kObjectInTheAreaEventType,
        "Object in the Area - prolonged event (caption) key2, STARTED",
        "Object in the Area - prolonged event (description) key2, STARTED",
        EventIsActive::yes,
        "key2"
    },
    {
        kObjectInTheAreaEventType,
        "Object in the Area - prolonged event (caption) key1, FINISHED",
        "Object in the Area - prolonged event (description) key1, FINISHED",
        EventIsActive::no,
        "key1"
    },
    {
        kObjectInTheAreaEventType,
        "Object in the Area - prolonged event (caption) NEW key1, STARTED",
        "Object in the Area - prolonged event (description) NEW key1, STARTED",
        EventIsActive::yes,
        "key1"
    },
    {
        kObjectInTheAreaEventType,
        "Object in the Area - prolonged event (caption) key2, FINISHED",
        "Object in the Area - prolonged event (description) key2, FINISHED",
        EventIsActive::no,
        "key2"
    },
    {
        kObjectInTheAreaEventType,
        "Object in the Area - prolonged event (caption) key1, FINISHED",
        "Object in the Area - prolonged event (description) key1, FINISHED",
        EventIsActive::no,
        "key1"
    },
    {
        kLineCrossingEventType,
        "Line crossing - impulse event (caption)",
        "Line crossing - impulse event (description)",
    },
    {
        kSuspiciousNoiseEventType,
        "Suspicious noise - group impulse event (caption)",
        "Suspicious noise - group impulse event (description)",
    },
    {
        kGunshotEventType,
        "Gunshot - group impulse event (caption)",
        "Gunshot - group impulse event (description)",
    },
    {
        kAdditionalEventType,
        "Caption: Additional Event",
        "Description: Additional Event",
    }
};

} // namespace

DeviceAgent::DeviceAgent(Engine* engine, const nx::sdk::IDeviceInfo* deviceInfo):
    ConsumingDeviceAgent(deviceInfo, NX_DEBUG_ENABLE_OUTPUT),
    m_engine(engine)
{
    m_eventThread = std::make_unique<std::thread>([this]() { processEvents(); });
}

DeviceAgent::~DeviceAgent()
{
    m_terminated = true;
    {
        std::unique_lock<std::mutex> lock(m_eventGenerationLoopMutex);
        m_eventGenerationLoopCondition.notify_all();
    }

    if (m_eventThread)
        m_eventThread->join();
}

/**
 * DeviceAgent manifest may declare eventTypes and objectTypes similarly to how an Engine declares
 * them - semantically the set from the Engine manifest is joined with the set from the DeviceAgent
 * manifest. Also this manifest should declare supportedEventTypeIds and supportedObjectTypeIds
 * lists which are treated as white-list filters for the respective set from Engine manifest
 * (absent lists are treated as empty lists, thus, disabling all types from the Engine).
 */
std::string DeviceAgent::manifestString() const
{
    static const std::string kAdditionalEventTypes = R"json(,
        {
            "id": ")json" + kAdditionalEventType + R"json(",
            "name": "Additional event 1"
        },
        {
             "id": ")json" + kAdditionalEventType2 + R"json(",
             "name": "Additional event 2"
        })json";

    std::string result = /*suppress newline*/ 1 + (const char*)R"json(
{
    "capabilities": "disableStreamSelection",
    "supportedEventTypeIds": [
        ")json" + kLineCrossingEventType + R"json(",
        ")json" + kSuspiciousNoiseEventType + R"json(",
        ")json" + kObjectInTheAreaEventType + R"json(",
        ")json" + kLoiteringEventType + R"json(",
        ")json" + kIntrusionEventType + R"json(",
        ")json" + kGunshotEventType + R"json(")json"
        + (m_deviceAgentSettings.declareAdditionalEventTypes ?
            ",\"" + kAdditionalEventType + "\", \"" + kAdditionalEventType2 + "\"" : "") + R"json(
    ],
    "typeLibrary": {
        "eventTypes": [
            {
                "id": ")json" + kLoiteringEventType + R"json(",
                "name": "Loitering"
            },
            {
                "id": ")json" + kIntrusionEventType + R"json(",
                "name": "Intrusion",
                "flags": "stateDependent|regionDependent"
            },
            {
                "id": ")json" + kGunshotEventType + R"json(",
                "name": "Gunshot",
                "groupId": ")json" + kSoundRelatedEventGroup + R"json("
            })json"
            + (m_deviceAgentSettings.declareAdditionalEventTypes ? kAdditionalEventTypes : "")
            + R"json(
        ]
    }
})json";

    return result;
}

Result<const ISettingsResponse*> DeviceAgent::settingsReceived()
{
    m_deviceAgentSettings.generateEvents = toBool(settingValue(kGenerateEventsSetting));

    m_deviceAgentSettings.declareAdditionalEventTypes =
        toBool(settingValue(kDeclareAdditionalEventTypesSetting));

    // The manifest depends on declareAdditionalEventTypes setting, so sending the new manifest.
    pushManifest(manifestString());

    return nullptr;
}

void DeviceAgent::doSetNeededMetadataTypes(
    Result<void>* /*outResult*/, const IMetadataTypes* neededMetadataTypes)
{
    if (neededMetadataTypes->isEmpty())
        stopFetchingMetadata();

    startFetchingMetadata(neededMetadataTypes);
}

void DeviceAgent::startFetchingMetadata(const IMetadataTypes* /*metadataTypes*/)
{
    std::unique_lock<std::mutex> lock(m_eventGenerationLoopMutex);
    NX_OUTPUT << __func__ << "() BEGIN";
    m_eventsNeeded = true;
    m_eventGenerationLoopCondition.notify_all();
    m_eventTypeId = kLineCrossingEventType; //< First event to produce.
    NX_OUTPUT << __func__ << "() END -> noError";
}

void DeviceAgent::stopFetchingMetadata()
{
    std::unique_lock<std::mutex> lock(m_eventGenerationLoopMutex);
    NX_OUTPUT << __func__ << "() BEGIN";
    m_eventsNeeded = false;
    NX_OUTPUT << __func__ << "() END -> noError";
}

void DeviceAgent::processEvents()
{
    while (!m_terminated)
    {
        if (m_deviceAgentSettings.generateEvents && m_eventsNeeded)
            pushMetadataPacket(cookSomeEvents());

        std::unique_lock<std::mutex> lock(m_eventGenerationLoopMutex);
        if (m_terminated)
            break;
        // Sleep until the next event needs to be generated, or the thread is ordered
        // to terminate (hence condition variable instead of sleep()). Return value
        // (whether the timeout has occurred) and spurious wake-ups are ignored.
        static const seconds kEventGenerationPeriod{3};
        m_eventGenerationLoopCondition.wait_for(lock, kEventGenerationPeriod);
    }
}

//-------------------------------------------------------------------------------------------------
// private

IMetadataPacket* DeviceAgent::cookSomeEvents()
{
    const auto descriptor = kEventsToFire[m_eventContext.currentEventTypeIndex];
    auto eventMetadataPacket = new EventMetadataPacket();
    eventMetadataPacket->setTimestampUs(usSinceEpoch());
    eventMetadataPacket->setDurationUs(0);

    auto eventMetadata = makePtr<EventMetadata>();
    eventMetadata->setTypeId(descriptor.eventTypeId);
    eventMetadata->setCaption(descriptor.caption);
    eventMetadata->setDescription(descriptor.description);
    eventMetadata->setIsActive(descriptor.isActive == EventIsActive::yes);
    eventMetadata->setKey(descriptor.prolongedEventKey);

    NX_PRINT/*NX_OUTPUT*/ << "Firing event: "
        << "type: " << eventMetadata->typeId()
        << ", isActive: " << eventMetadata->isActive();

    m_eventContext.currentEventTypeIndex =
        (m_eventContext.currentEventTypeIndex + 1) % ((int) kEventsToFire.size());

    eventMetadataPacket->addItem(eventMetadata.get());
    return eventMetadataPacket;
}

int64_t DeviceAgent::usSinceEpoch() const
{
    return duration_cast<microseconds>(
        system_clock::now().time_since_epoch()).count();
}

} // namespace events
} // namespace stub
} // namespace analytics
} // namespace vms_server_plugins
} // namespace nx
